import os
os.environ["CUDA_VISIBLE_DEVICES"] = "0"

# tensorflow 2.4.0
import tensorflow
print(tensorflow.__version__)

import tensorflow as tf
print("Num GPUs Available: ", len(tf.config.list_physical_devices('GPU')))

from tensorflow.keras import Input
from tensorflow.keras.layers import Conv2D, Flatten, Dense, Conv2DTranspose, Reshape, Lambda, Activation, BatchNormalization, LeakyReLU, Dropout, ZeroPadding2D, UpSampling2D, Embedding, multiply
from tensorflow.keras.models import Model, Sequential
from tensorflow.keras.optimizers import Adam, RMSprop
from tensorflow.keras.initializers import RandomNormal
from tensorflow.keras.preprocessing import image_dataset_from_directory

# Commented out IPython magic to ensure Python compatibility.
import numpy as np
import pandas as pd
import os
import matplotlib
import matplotlib.pyplot as plt
# %matplotlib inline

from sklearn.metrics import confusion_matrix

BATCH_SIZE = 61
IMAGE_SIZE_H = 80
IMAGE_SIZE_W = 128

generator_initial_dense_layer_size = (5, 8, 512)
generator_upsample = [1,1,1,1]
generator_conv_filters = [256,128,64,1]
generator_conv_kernel_size = [3,3,3,3]
generator_conv_strides = [2,2,2,2]
generator_batch_norm_momentum = 0.9
generator_activation = 'leaky_relu'
generator_dropout_rate = None
generator_learning_rate = 0.0002

z_dim = 100
batch_size = BATCH_SIZE
num_classes = 4

n_layers_generator = len(generator_conv_filters)

weight_init = RandomNormal(mean=0., stddev=0.02)
epoch_counter = 0

def _build_generator():
    latent = Input(shape=(z_dim,), name='z_dim__generator_input')
    image_class = Input(shape=(1,), dtype='int32', name='image_class__generator_input')
    
    classes = Embedding(num_classes, z_dim,
                    embeddings_initializer='glorot_normal')(image_class)
    h = multiply([latent, classes])
    x = h
    x = Dense(np.prod(generator_initial_dense_layer_size), kernel_initializer = weight_init)(x)
    if generator_batch_norm_momentum:
        x = BatchNormalization(momentum = generator_batch_norm_momentum)(x)
    x = get_activation(generator_activation)(x)
    x = Reshape(generator_initial_dense_layer_size)(x)

    if generator_dropout_rate:
        x = Dropout(rate = generator_dropout_rate)(x)

    for i in range(n_layers_generator):
        if generator_upsample[i] == 2:
            x = UpSampling2D()(x)
            x = Conv2D(
            filters = generator_conv_filters[i]
            , kernel_size = generator_conv_kernel_size[i]
            , padding = 'same'
            , name = 'generator_conv_' + str(i)
            , kernel_initializer = weight_init
            )(x)
        else:
            x = Conv2DTranspose(
                filters = generator_conv_filters[i]
                , kernel_size = generator_conv_kernel_size[i]
                , padding = 'same'
                , strides = generator_conv_strides[i]
                , name = 'generator_conv_' + str(i)
                , kernel_initializer = weight_init
                )(x)

        if i < n_layers_generator - 1:
            if generator_batch_norm_momentum:
                x = BatchNormalization(momentum = generator_batch_norm_momentum)(x)
            x = get_activation(generator_activation)(x)
        else:
            x = Activation('tanh')(x)

    generator_output = x
    generator = Model([latent, image_class], generator_output, name='generator')
    generator.summary()
    return generator

def get_activation(activation):
    if activation == 'leaky_relu':
        layer = LeakyReLU(alpha = 0.2)
    else:
        layer = Activation(activation)
    return layer

generator = _build_generator()

generator.load_weights('trained_model/first_fold_generator-2598.h5')

generator.summary()

def cal_acc(my_confusion_matrix):
    top1_sum = 0
    top2_sum = 0
    for i in range(0,4):
        for j in range(0,4):
            if i == j:
                top1_sum = top1_sum + my_confusion_matrix[i][j]
                top2_sum = top2_sum + my_confusion_matrix[i][j]
            if i+1 == j:
                top2_sum = top2_sum + my_confusion_matrix[i][j]
    return top1_sum/61, top2_sum/61

def generate_img(generate_num=10):
    cwd = os.getcwd()
    print("cwd:", cwd)
    PATH = "fake_data"
    os.system("rm -rf %s" % (PATH))
    
    stages = ["c0", "c1", "c2", "c3"]
    if not os.path.exists(PATH):
        print("PATH:", PATH)
        os.makedirs(PATH)

    for i in range(0, len(stages)):
        stage_PATH = os.path.join(PATH, stages[i])
        print("stage_PATH: %s" % (stage_PATH))
        if not os.path.exists(stage_PATH):
            print('makedirs("%s")' % (stage_PATH))
            os.makedirs(stage_PATH)
        noise = np.random.normal(0, 1, (generate_num, z_dim))
        zeros = np.zeros(generate_num) + i
        gen_imgs = generator.predict([noise, zeros])
        gen_imgs = gen_imgs.reshape((generate_num, 80, 128))
        gen_imgs = (gen_imgs+1)/2
        for j in range(0, len(gen_imgs)):
            file_name = ('%s-%d.jpg' % (stages[i],j))
            file_name = os.path.join(stage_PATH, file_name)
            matplotlib.image.imsave(file_name, gen_imgs[j],vmin = 0, vmax = 1, cmap='gray')
            
def plot_training_curve(train_list, test_list):
    train_list = np.array(train_list)
    test_list = np.array(test_list)

    acc = train_list[:, 1]
    val_acc = test_list[:, 1]
    loss = train_list[:, 0]
    val_loss = test_list[:, 0]

    plt.figure(figsize=(8, 8))
    plt.subplot(2, 1, 1)
    plt.plot(acc, label='Training Accuracy')
    plt.plot(val_acc, label='Validation Accuracy')
    plt.ylim([0.4, 1])
    plt.legend(loc='lower right')
    plt.title('Training and Validation Accuracy')

    plt.subplot(2, 1, 2)
    plt.plot(loss, label='Training Loss')
    plt.plot(val_loss, label='Validation Loss')
    plt.ylim([0, 2.0])
    plt.legend(loc='upper right')
    plt.title('Training and Validation Loss')
    plt.xlabel('epoch')
    plt.show()

def fit(generate_num, synData_percentage, fine_tune_epochs=1000):
    Number_of_images = 607
    IMG_SIZE = (80, 128)
    syn_BATCH_SIZE = int(Number_of_images*synData_percentage)
    print("generate_num=", generate_num, ", synData_percentage=", synData_percentage, ", syn_BATCH_SIZE=", syn_BATCH_SIZE)

    print("Generate Fake Images")
    generate_img(generate_num)

    fold = "celeb_gray_"
    PATH_train = os.path.join(os.getcwd(), fold+'train')
    PATH_synTrain = os.path.join(os.getcwd(), 'fake_data')
    PATH_test = os.path.join(os.getcwd(), fold+'test')
    print("Load dataset")
    train_dataset = image_dataset_from_directory(PATH_train,
                                                 label_mode='categorical',
                                                 shuffle=True,
                                                 batch_size=Number_of_images,
                                                 image_size=IMG_SIZE, 
                                                 seed=2020,
                                                )

    synTrain_dataset = image_dataset_from_directory(PATH_synTrain,
                                                     label_mode='categorical',
                                                     shuffle=True,
                                                     batch_size=syn_BATCH_SIZE,
                                                     image_size=IMG_SIZE, 
                                                     seed=2020,
                                                   )
    validation_dataset = image_dataset_from_directory(PATH_test,
                                                      label_mode='categorical',
                                                      shuffle=True,
                                                      batch_size=batch_size,
                                                      image_size=IMG_SIZE,
                                                      seed=2020,
                                                     )
    train_img, train_label = next(iter(train_dataset))
    print(train_img.shape, train_label.shape)
    synTrain_img, synTrain_label = next(iter(synTrain_dataset))
    print(synTrain_img.shape, synTrain_label.shape)
    validation_img, validation_label = next(iter(validation_dataset))
    print(validation_img.shape, validation_label.shape)

    AUTOTUNE = tf.data.experimental.AUTOTUNE
    train_dataset = train_dataset.prefetch(buffer_size=AUTOTUNE)
    validation_dataset = validation_dataset.prefetch(buffer_size=AUTOTUNE)

    data_augmentation = tf.keras.Sequential([
      tf.keras.layers.experimental.preprocessing.RandomFlip('horizontal'),
    ])

    print("PLT train_dataset")
    for image, _ in train_dataset.take(1):
      plt.figure(figsize=(10, 10))
      first_image = image[0]
      for i in range(9):
        ax = plt.subplot(3, 3, i + 1)
        augmented_image = data_augmentation(tf.expand_dims(first_image, 0))
        plt.imshow(augmented_image[0] / 255)
        plt.axis('off')
      plt.show()

    preprocess_input = tf.keras.applications.mobilenet_v2.preprocess_input
    rescale = tf.keras.layers.experimental.preprocessing.Rescaling(1./127.5, offset= -1)

    IMG_SHAPE = IMG_SIZE + (3,)
    
    from datetime import datetime, timezone, timedelta
    tz = timezone(timedelta(hours=+8))
    now = datetime.now(tz)
    str_now = now.strftime("%b_%d_%Y_%H_%M_%S")
    base_model_name = "MobileNetV2_" + fold + str_now
    print("Load MobileNetV2 Model")
    base_model = tf.keras.applications.MobileNetV2(input_shape=IMG_SHAPE,
                                                   include_top=False,
                                                   weights='imagenet')

    image_batch, label_batch = next(iter(train_dataset))
    feature_batch = base_model(image_batch)

    base_model.trainable = False
    global_average_layer = tf.keras.layers.GlobalAveragePooling2D()
    feature_batch_average = global_average_layer(feature_batch)

    prediction_layer = tf.keras.layers.Dense(4, activation='softmax')
    prediction_batch = prediction_layer(feature_batch_average)

    inputs = tf.keras.Input(shape=IMG_SHAPE)
    x = data_augmentation(inputs)
    x = preprocess_input(x)
    x = base_model(x, training=False)
    x = global_average_layer(x)
    x = tf.keras.layers.Dropout(0.4)(x)
    outputs = prediction_layer(x)
    model = tf.keras.Model(inputs, outputs)

    base_learning_rate = 0.0002
    model.compile(optimizer=tf.keras.optimizers.Adam(lr=base_learning_rate),
                  loss='categorical_crossentropy',
                  metrics=['accuracy'])
    model.summary()


    train_img, train_label = next(iter(train_dataset))
    synTrain_img, synTrain_label = next(iter(synTrain_dataset))
    train_img = np.concatenate((train_img, synTrain_img), axis=0)
    train_label = np.concatenate((train_label, synTrain_label), axis=0)

    train_list = []
    test_list = []
    epoch_list = []
    val_label_list = []
    model_predict_list = []

    base_model.trainable = True
    print("Number of layers in the base model: ", len(base_model.layers))
    fine_tune_at = 100
    for layer in base_model.layers[:fine_tune_at]:
        layer.trainable =  False
    model.compile(loss='categorical_crossentropy',
                  optimizer = tf.keras.optimizers.RMSprop(lr=base_learning_rate/10),
                  metrics=['accuracy'])
    model.summary()

    train_list = []
    test_list = []
    top2acc_list = []
    epochs = 0
    val_img, val_label = next(iter(validation_dataset))
    for i in range(epochs, fine_tune_epochs):
        print("epoch:", epochs)
        train_img, train_label = next(iter(train_dataset))
        synTrain_img, synTrain_label = next(iter(synTrain_dataset))
        train_img = np.concatenate((train_img, synTrain_img), axis=0)
        train_label = np.concatenate((train_label, synTrain_label), axis=0)
        train = model.train_on_batch(train_img, train_label)
        val_e = model.evaluate(val_img, val_label)
        val_label_1D = np.argmax(val_label, axis=1)
        print(val_label_1D)

        model_predict = model.predict(val_img)
        model_predict_1D = np.argmax(model_predict, axis=1)
        print(model_predict_1D)

        my_confusion_matrix = confusion_matrix(val_label_1D, model_predict_1D)
        top1acc, top2_acc = cal_acc(my_confusion_matrix)
        print("top1acc:", top1acc, "top2acc:", top2_acc)
        
        if val_e[1] >= 0.75:
            weights_PATH = base_model_name + "_weights"
            if not os.path.exists(weights_PATH):
                os.makedirs(weights_PATH)
            weights_PATH = os.path.join(weights_PATH, 'transfer_learning_g' + str(generate_num) + '_p' + str(synData_percentage) +'weights-%d.h5' % (epochs))
            model.save_weights(weights_PATH)
            
        epoch_list.append(epochs)
        train_list.append(train)
        test_list.append(val_e)
        top2acc_list.append(top2_acc)
        val_label_list.append(val_label_1D)
        model_predict_list.append(model_predict_1D)
        epochs = epochs + 1

    generate_num_list.append(generate_num)
    syn_BATCH_SIZE_list.append(syn_BATCH_SIZE)
    max_valacc = max(np.array(np.array(test_list)[:, 1]))
    max_valacc_list.append(max_valacc)
    max_top2acc = max(np.array(top2acc_list))
    print("generate_num=", generate_num, ", synData_percentage=", synData_percentage, ", max_valacc=", max_valacc, ", max_top2acc=", max_top2acc)
    
    plot_training_curve(np.array(train_list), np.array(test_list))
    
    df = pd.DataFrame({'epoch_list': epoch_list, 'train_list': train_list, 'test_list': test_list, 'top2acc_list': top2acc_list, "val_label_list":val_label_list, 'model_predict_list':model_predict_list})
    acc_PATH = base_model_name + "_acc"
    if not os.path.exists(acc_PATH):
        os.makedirs(acc_PATH)
    acc_PATH = os.path.join(acc_PATH, 'out_g' + str(generate_num) + '_p' + str(synData_percentage) + '.csv')
    df.to_csv(acc_PATH, index=False)

generate_num_list = []
syn_BATCH_SIZE_list = []
max_valacc_list = []
generate_num_list = [10, 100, 1000, 10000]
synData_percentage_list = [0.1, 0.5, 1, 10]
fine_tune_epochs = 1000
for i in range(0, len(generate_num_list)):
    for j in range(0, len(synData_percentage_list)):
        fit(generate_num_list[i], synData_percentage_list[j], fine_tune_epochs)